/***************************************************************************
 *   Copyright (C) 2010 by azrael,,,   *
 *   kzhevanv@rambler.ru   *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/


#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "hash-table.h"

#ifndef FALSE
#define FALSE (0)
#endif

#ifndef TRUE
#define TRUE (!FALSE)
#endif


enum sex{
	SEX_FEMALE = 0,
	SEX_MALE = 1,
	SEX_ALIEN = 2
};

struct user_data{
	char *name;
	int age;
	int sex;
};




static const unsigned int prime[] = {
5,
11,
19,
37,
73,
109,
163,
251,
367,
557,
823,
1237,
1861,
2777,
4177,
6247,
9371,
14057,
21089,
31627,
47431,
71143,
106721,
160073,
240101,
360163,
540217,
810343,
1215497,
1823231,
2734867,
4102283,
6153409,
9230113,
13845163
};


static unsigned int default_hash (void *key)
{
	return (unsigned int) key;
}

static int default_compare (const void *a,const void *b)
{
	if (a > b)
		return 1;
	else if (a < b)
		return -1;
	return 0;
}

unsigned int djb_strhash(void *key)
{
	unsigned int hash = 5381;
	unsigned char *str = (unsigned char *) key;
	int c;
 
	while ((c = *str++) != 0)
		hash = ((hash << 5) + hash) + c;
 
	return hash;
}

struct hash_table *hash_table_new(size_t size_hint, hash_callback_t hash, hash_compare_t keycmp)
{
	if (hash == NULL) 
		hash = *default_hash;
	if (keycmp == NULL)
		keycmp = *default_compare;

	struct hash_table *t;

	t = malloc(sizeof(struct hash_table));
	if (t == NULL)
		return NULL;

	t->table = malloc(sizeof(struct hash_bucket) * size_hint);
	if (t->table == NULL) {
		free(t);
		return NULL;
	}
	
	memset(t->table, 0, sizeof(struct hash_bucket));
	t->compare = keycmp;
	t->hash = hash;
	t->size = size_hint;
	t->count = 0;
	t->collisions = 0;
	
	return t;
}

int hash_table_insert(struct hash_table *table, void *key, void *data)
{
	unsigned int idx;
	float coef;	
	
	if (table == NULL)
		return -1;
	
	coef = (float)table->count / (float)table->collisions;
	if (coef <= 2.0)
		hash_rebuild(table);
	
	idx = (*table->hash)(key) % table->size;

	if (table->table[idx].key == NULL) {
		table->table[idx].key = key;
		table->table[idx].value = data;
	} else {
		struct hash_bucket *b;
		b = malloc(sizeof(*b));
		if (b == NULL)
			return -1;
		memset(b, 0, sizeof(*b));
		b->key = key;
		b->value = data;
		b->next = table->table[idx].next;
		table->table[idx].next = b;
		++table->collisions;
	}

	++table->count;

	return 0;
}

boolean_t hash_table_remove(struct hash_table *table, void *key)
{
	struct hash_bucket *cur;
	struct hash_bucket *prev;
	unsigned int index;

	prev = NULL;
	index = table->hash(key) % (table->size);

	cur = &table->table[index];
	
	if (cur->key == NULL)
		return  FALSE;

	while (cur) {
		if (table->compare(key, cur->key) == 0)
			break;
		prev = cur;
		cur = cur->next;
	}

	if (cur == NULL)
		return FALSE;

	if (prev == NULL) {
		if (cur->next != NULL){
			table->table[index].key = cur->next->key;
			table->table[index].value = cur->next->value;
			table->table[index].next = cur->next->next;
			--table->collisions;
		} else {
			table->table[index].key = NULL;
			table->table[index].value = NULL;
		}
	} else {
		prev->next = cur->next;
		free(cur);
		--table->collisions;
	}
	--table->count;
	
	return TRUE;
}

void hash_table_destroy(struct hash_table **table)
{
	struct hash_bucket *cur, *next, *prev;
	unsigned int idx;
	
	if ((table == NULL) && (*table == NULL))
		return;

	for (idx = 0; idx < (*table)->size; ++idx) {
		cur = &(*table)->table[idx];
		prev = &(*table)->table[idx];
		for (cur = cur->next; cur != NULL; cur = next) {
			next = cur->next;
			prev->next = cur->next;
			free(cur);
		}
	}
	free((*table)->table);
	free(*table);
	*table = NULL;
}

boolean_t hash_table_replace(struct hash_table *table, void *key, void *data)
{
	unsigned int idx;

	if (table == NULL)
		return -1;

	idx = (*table->hash)(key) % table->size;

	struct hash_bucket *cur;

	for (cur = &table->table[idx]; cur != NULL; cur->next) {
		if ((table->compare)(cur->key, key) == 0)
			break;
	}

	if (cur == NULL)
		return -1;

	cur->value = data;
}

struct hash_table_iter *hash_table_iterate_init(struct hash_table *table)
{
	struct hash_table_iter *i;
	
	i = malloc(sizeof(*i));
	if (i == NULL)
		return NULL;
	
	i->table = table;
	i->index = 0;
	i->bucket = &table->table[0];
	return i;
}

void hash_table_iterate_deinit(struct hash_table_iter **iter) 
{
	struct hash_bucket *cur, *next, *prev;
	unsigned int idx;
	
	(*iter)->table = NULL;
	(*iter)->index = 0;
	(*iter)->bucket = NULL;
	
	free((*iter)->bucket);
	free(*iter);	
}

boolean_t hash_table_iterate(struct hash_table_iter *iter, void **res_key, void **res_data)
{
	int i;
	
	if (iter == NULL)
		return -1;
	
	i = iter->index;
	
	while (i < iter->table->size) {
		if (iter->bucket->key == NULL) {
			++i;
			iter->bucket = &iter->table->table[i];
		} else {
			iter->index = i;
			break;
		}
	}
	
	if (i < iter->table->size) {
	
		if (res_key != NULL)
			*res_key = iter->bucket->key;
	
		if (res_data != NULL)
			*res_data = iter->bucket->value;

		if (iter->bucket->next != NULL) {
			iter->bucket = iter->bucket->next;
		} else {
			++i;
			if (i < iter->table->size) {
				iter->bucket = &iter->table->table[i];
				iter->index = i;
			} else {
				iter->bucket = NULL;
				iter->index = i;
			}
		}
		return TRUE;
	}
	return FALSE;
}

int hash_rebuild(struct hash_table *hash)
{
	size_t new_size;
	struct hash_bucket *new_table;
	struct hash_bucket *cur, *next;
	int idx;
	int index;
	int i;
	unsigned int collisions;
	
	collisions = 0;
	
	for (i = 0; i < 36 ; i++) {
		if (hash->size < prime[i]) {
			new_size = prime[i];
			break;
		}
	}
	
	new_table = malloc(sizeof(*new_table) * new_size);
	if (new_table == NULL)
		return -1;

	memset(new_table, 0, sizeof(*new_table) * new_size);
	for (idx = 0; idx < hash->size; idx++) {
		cur = &(hash->table[idx]);
		
		if (cur->key == NULL)
			continue;
		
		index = (*hash->hash)(cur->key) % new_size;
		
		if (new_table[index].key == NULL) {
			new_table[index].key = cur->key;
			new_table[index].value = cur->value;
		} else {
			struct hash_bucket *b;
			b = malloc(sizeof(*b));
			if (b == NULL) {
				free(new_table);
				return -1;
			}
			memset(b, 0, sizeof(*b));
			b->key = cur->key;
			b->value = cur->value;
			b->next = new_table[index].next;
			new_table[index].next = b;
			++collisions;
		}
		for (cur = cur->next; cur != NULL; cur = next) {
			next = cur->next;
			index = (*hash->hash)(cur->key) % new_size;
			if (new_table[index].key == NULL) {
				new_table[index].key = cur->key;
				new_table[index].value = cur->value;
				free(cur);
			} else {
				cur->next = new_table[index].next;
				new_table[index].next = cur;
				++collisions;
			}
		}
		free(cur);
	}
	hash->table = new_table;	
	hash->collisions = collisions;
	hash->size = new_size;

	return 0;
}

int hash_merge(struct hash_table *table_to, struct hash_table *table_from)
{
	int idx;
	int index;
	int t;
	struct hash_bucket *cur, *next;
	
	for (idx = 0; idx < table_from->size; idx++) {
		cur = &table_from->table[idx];
		if (cur->key == NULL)
			continue;
		index = (*table_to->hash)(cur->key) % table_to->size;
		if (table_to->table[index].key == NULL) {
			table_to->table[index].key = cur->key;
			table_to->table[index].value = cur->value;
		} else {
			struct hash_bucket *b;
			b = malloc(sizeof(*b));
			if (b == NULL) {
				return -1;
			}
			memset(b, 0, sizeof(*b));
			b->key = cur->key;
			b->value = cur->value;
			b->next = table_to->table[index].next;
			table_to->table[index].next = b;
			++table_to->collisions;
		}
		for (cur = cur->next; cur != NULL; cur = next) {
			next = cur->next;
			index = (*table_to->hash)(cur->key) % table_to->size;
			if (table_to->table[index].key == NULL) {
				table_to->table[index].key = cur->key;
				table_to->table[index].value = cur->value;
			} else {
				struct hash_bucket *b;
				b = malloc(sizeof(*b));
				if (b == NULL) {
					return -1;
				}
				memset(b, 0, sizeof(*b));
				b->key = cur->key;
				b->value = cur->value;
				b->next = table_to->table[index].next;
				table_to->table[index].next = b;
				++table_to->collisions;
			}
		}
	}
	table_to->count = table_from->count + table_to->count;
}

struct male_human {
	char *name;
	int age;
};
 
struct male_human male_table[] = {
	{ "Sergey", 20 },
	{ "AMD", 32},
	{ "Denis", 21 },
	{ "intel", 64},
	{ "Alexey", 20},
	{ NULL, -1 }
};

int main(int argc, char *argv[])
{
	struct user_data *up, *up2;
	struct hash_table *table;
	struct hash_table *table2;
	struct male_human *hi;

	table = hash_table_new(1, djb_strhash, NULL);
	table2 = hash_table_new(1, djb_strhash, NULL);

	for (hi = male_table; hi->name != NULL; hi++) {
		up = malloc(sizeof(*up));
		/*strncpy(up->name, hi->name, sizeof(up->name));
		up->name[sizeof(up->name) - 1]='\0';*/
		up->name = hi->name;
		up->age = hi->age;
		up->sex = SEX_MALE;
		hash_table_insert(table, up->name, up);
	}
	up2 = malloc(sizeof(*up2));
	up2->name = "AMD";
	up2->age = 72;
	up2->sex = SEX_MALE;

	/*hash_table_remove(table, "Denis"); */
	
	hash_table_replace(table, up2->name, up2 );


	struct user_data *u;
	struct hash_table_iter *iter;
	iter = hash_table_iterate_init(table);
	/*hash_table_iterate_deinit(&iter);
	iter = hash_table_iterate_init(table);*/

	while(hash_table_iterate(iter, NULL, (void *) &u)){
		printf("Name: %s\n",u->name);
		printf("Age: %d\n",u->age);
	}
	/*printf("table2 ");
	hash_merge(table2, table);
	iter = hash_table_iterate_init(table2);
	while(hash_table_iterate(iter, NULL, (void *) &u)){
	printf("Name: %s\n",u->name);
	printf("Age: %d\n",u->age);
	}*/

	

	/*for (hi = male_table; hi->name != NULL; hi++) {
		hash_table_remove(table, hi->name);
	} */
	
	hash_table_iterate_deinit(&iter);

	hash_table_destroy(&table);

	return EXIT_SUCCESS;
}
